// CIS534 homework
// Author: Prof. Milo Martin
// University of Pennsylvania
// Spring 2010

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <unistd.h>
#include <getopt.h>
#include <unistd.h>
#include <assert.h>
#include <time.h>
#include <ctime>
#include <iostream>
#include <set>

#include "ppp.h"
#include "barrier.h"
#include "compute.h"
#include "random.h"

static int64_t distance(int64_t a, int64_t b);
int64_t check_one(int64_t particle, int64_t num_particles, int64_t *location, int64_t *weight, int64_t *radius);
int64_t check_two(int64_t particle, int64_t num_particles, int64_t *location, int64_t *weight, int64_t *radius);
int64_t check_three(int64_t particle, int64_t num_particles, int64_t *xlocation, int64_t *ylocation, int64_t *weight, int64_t radius);

const char* gethost()
{
  const char* host = getenv("HOST");
  if (host == NULL) {
    host = "<unknown>";
  }
  return host;
}

const char* getuser()
{
  const char* user = getenv("USER");
  if (user == NULL) {
    user = "<unknown>";
  }
  return user;
}

int64_t bigrand(int64_t max, int dist=1)
{
  if (max < dist) {
    dist = 1;
  }
  int64_t range = max/dist;
  int64_t rand = 0;
  for (int i=0; i<dist; i++) {
    int64_t r = (ppp::threadsafe_random() << 31) ^ ppp::threadsafe_random();
    rand += (r % range);
  }
  assert(rand < max);
  return rand;
}

double get_seconds()
{
  struct timespec ts;
  clock_gettime(CLOCK_REALTIME, &ts);
  double tv_sec = double(ts.tv_sec);
  double tv_nsec = double(ts.tv_nsec);
  double seconds = tv_sec + (tv_nsec / double(1000000000.0));
  return seconds;
}

void fill_rand(int64_t n, int64_t* array, int max_value)
{
  // Floyd Algorithm for generating a set of unique random numbers
  // Programming Pearls, p. 225
  // Note: numbers are unique, but would need to be shuffled to
  // guarantee the order is random, too

  std::set<int64_t> S;
  int64_t j = max_value - n;
  for (int64_t i = 0; i<n; i++) {
    int64_t t = bigrand(j+1, 3);
    if (S.find(t) != S.end()) { // If already present, just use "j"
      t = j;
    }
    j++;
    S.insert(t);
    array[i] = t;
  }
}

// Parameters and default values
static int64_t num_particles = 0;
static int64_t sparsity = 16;
static int64_t max_weight = 1000;
static int64_t radius = 500;
static int compute_id = 0;
static int num_trials = 1;
static int num_threads = 1;

// Data arrays
static int64_t* xlocation_array;
static int64_t* ylocation_array;
static int64_t* weight_array;
static int64_t* radius_array;
static int64_t* answer_array;
static int64_t* check_particle_array;
static int64_t* check_answer_array;


void driver()  // Called for each thread
{
  const int num_to_check = 100;
  // Calculated values
  int64_t max_location = num_particles * sparsity;
  int64_t half_radius = int64_t((double(radius)/2) * double(sparsity));
  if (compute_id == 3) {
    half_radius *= sparsity; // Enlarge the radius for 2D version
  }

  if (ppp::get_thread_id() == 0) {

    // Allocation 
    xlocation_array = new int64_t[num_particles];
    ylocation_array = new int64_t[num_particles];
    weight_array = new int64_t[num_particles];
    radius_array = new int64_t[num_particles];
    answer_array = new int64_t[num_particles];

    check_particle_array = new int64_t[num_to_check];
    check_answer_array = new int64_t[num_to_check];
  }

  double elapsed_time = 0.0;
  for (int trial = 1; trial <= num_trials; trial++) {

    // Initialization
    if (ppp::get_thread_id() == 0) {

      fill_rand(num_particles, xlocation_array, max_location);
      if (compute_id == 3) {
        fill_rand(num_particles, ylocation_array, max_location);
      }
    
      for (int64_t i = 0; i<num_particles; i++) {
        weight_array[i] = bigrand(max_weight);
        radius_array[i] = half_radius + bigrand(half_radius);
        answer_array[i] = 0;
      }
      
      // Pre-calculate values for randomized spot checking
      for (int64_t i = 0; i<num_to_check; i++) {
        int64_t particle_to_check = bigrand(num_particles);
        check_particle_array[i] = particle_to_check;
        if (compute_id == 1) {
          check_answer_array[i] = check_one(particle_to_check, num_particles, xlocation_array, weight_array, radius_array);
        } else if (compute_id == 2) {
          check_answer_array[i] = check_two(particle_to_check, num_particles, xlocation_array, weight_array, radius_array);
        } else if (compute_id == 3) {
          check_answer_array[i] = check_three(particle_to_check, num_particles, xlocation_array, ylocation_array, weight_array, int64_t(half_radius*1.5));
        } else {
          check_answer_array[i] = 0;
        }
      }
    }

    double start_time = get_seconds();
    
    ppp::barrier();
    if (compute_id == 1) {
      compute_one(num_particles, xlocation_array, weight_array, radius_array, answer_array);
    } else if (compute_id == 2) {
      compute_two(num_particles, xlocation_array, weight_array, radius_array, answer_array);
    } else if (compute_id == 3) {
      compute_three(num_particles, xlocation_array, ylocation_array, weight_array, int64_t(half_radius*1.5), answer_array);
    }
    ppp::barrier();
    
    double end_time = get_seconds();
    double seconds = end_time - start_time;
    
    if (ppp::get_thread_id() == 0) {
      // Check the pre-calculated answers for randomized spot checking
      for (int64_t i = 0; i<num_to_check; i++) {
        int64_t particle_to_check = check_particle_array[i];
        int64_t expected_answer = check_answer_array[i];
        int64_t calculated_answer = answer_array[particle_to_check];
        if (expected_answer != calculated_answer) {
          std::cout << "ERROR: particle " << particle_to_check << " had value `" << calculated_answer
                    << "` not the expected value `" << expected_answer << "`\n";
        }
      }
    }
    
    if (ppp::get_thread_id() == 0) {
      std::cout << "trial: " << trial << ", ";
      std::cout << "user: " << getuser() << ", ";
      std::cout << "host: " << gethost() << ", ";
      std::cout << "threads: " << num_threads << ", ";
      std::cout << "comp: " << compute_id << ", ";
      std::cout << "particles: " << num_particles << ", ";
      //      std::cout << "radius: " << radius << ", ";
      //      std::cout << "sparsity: " << sparsity << ", ";
      std::cout << "seconds: " << seconds << "\n";
    }
    
    elapsed_time += seconds;
  }
  
  if (ppp::get_thread_id() == 0) {
    std::cout << "average:  ";
    std::cout << "user: " << getuser() << ", ";
    std::cout << "host: " << gethost() << ", ";
    std::cout << "threads: " << num_threads << ", ";
    std::cout << "comp: " << compute_id << ", ";
    std::cout << "particles: " << num_particles << ", ";
    //    std::cout << "radius: " << radius << ", ";
    //    std::cout << "sparsity: " << sparsity << ", ";
    std::cout << "seconds: " << elapsed_time/double(num_trials) << "\n";
  }
  
}

static struct option long_options[] =
  {
    {"particles", required_argument, 0, 'p'},
    {"sparsity", required_argument, 0, 's'},
    {"weight", required_argument, 0, 'w'},
    {"radius", required_argument, 0, 'r'},
    {"computation", required_argument, 0, 'c'},
    {"trials", required_argument, 0, 't'},
    {"threads", required_argument, 0, 'n'},
    {0, 0, 0, 0}
  };

int main(int argc, char** argv)
{
  
  // Code based on the example code in the getopt documentation
  while (true) {

    int option_index = 0;    
    int c = getopt_long_only(argc, argv, "p:s:w:r:c:t:",
                             long_options, &option_index);
    
    /* Detect the end of the options. */
    if (c == -1)
      break;
    
    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      break;
      
    case 'p':
      num_particles = atoll(optarg);
      break;
      
    case 's':
      sparsity = atoll(optarg);
      break;
      
    case 'w':
      max_weight = atoll(optarg);
      break;
      
    case 'r':
      radius = atoll(optarg);
      break;
      
    case 'c':
      compute_id = atoi(optarg);
      break;

    case 't':
      num_trials = atoi(optarg);
      break;

    case 'n':
      num_threads = atoi(optarg);
      break;
      
    case '?':
      /* getopt_long already printed an error message. */
      exit(1);
      break;
      
    default:
      exit(1);
    }
  }

  if ((compute_id <= 0) || (compute_id >= 4)) {
    fprintf(stderr, "Error: --computation must be set to 1, 2, or 3\n");
    exit(1);
  }

  if (num_particles <= 0) {
    fprintf(stderr, "Error: --particles must be set\n");
    exit(1);
  }
  
  ppp::init(driver, num_threads);
}



/******************** code for checking the results *********************/

int64_t check_one(int64_t particle, int64_t num_particles, int64_t *location, int64_t *weight, int64_t *radius)
{
  // Computation 1
  int64_t answer = 0;
  for (int64_t j = 0; j < num_particles; j++) {
    if (distance(location[particle], location[j]) < radius[particle]) {
       answer += weight[j];
    }
  }
  return answer;
}

int64_t check_two(int64_t particle, int64_t num_particles, int64_t *location, int64_t *weight, int64_t *radius)
{
  // Computation 2
  int64_t answer = 0;
  for (int64_t j = 0; j < num_particles; j++) {
    if (distance(location[particle], location[j]) < radius[particle]) {
      answer += (weight[particle] + weight[j]) | distance(location[particle], location[j]);
    }
  }
  return answer;
}

int64_t check_three(int64_t particle, int64_t num_particles, int64_t *xlocation, int64_t *ylocation, int64_t *weight, int64_t radius)
{
  // Computation 3
  int64_t answer = 0;
  for (int64_t j = 0; j < num_particles; j++) {
    int64_t dist = distance(xlocation[particle], xlocation[j]) + distance(ylocation[particle], ylocation[j]);
    if (dist < radius) {
      answer += (weight[particle] + weight[j]) * dist;
    }
  }
  return answer;
}

static int64_t distance(int64_t a, int64_t b) 
{
  if (a > b) {
    return a-b;
  } else {
    return b-a;
  }
}
