// CIS534 homework
// Author: Prof. Milo Martin
// University of Pennsylvania
// Spring 2010

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <unistd.h>
#include <getopt.h>
#include <unistd.h>
#include <assert.h>
#include <time.h>
#include <ctime>
#include <iostream>
#include <set>

#include "ppp.h"
#include "barrier.h"
#include "random.h"

#include "TTSLock.h"
#include "TicketLock.h"
#include "ArrayLock.h"
#include "QueueLock.h"

const char* gethost()
{
  const char* host = getenv("HOST");
  if (host == NULL) {
    host = "<unknown>";
  }
  return host;
}

const char* getuser()
{
  const char* user = getenv("USER");
  if (user == NULL) {
    user = "<unknown>";
  }
  return user;
}

double get_seconds()
{
  struct timespec ts;
  clock_gettime(CLOCK_REALTIME, &ts);
  double tv_sec = double(ts.tv_sec);
  double tv_nsec = double(ts.tv_nsec);
  double seconds = tv_sec + (tv_nsec / double(1000000000.0));
  return seconds;
}

// Parameters and default values
static int num_threads = 1;
static int num_iterations = 0;
static int num_counters = 1024*16;
static int num_locks = 0;
static int hold_time = 250;
static int think_time = 250;

static ppp::TTSLock check_lock;

// A typedef to specify the type of lock we're using
#ifndef LOCKTYPE
#error "LOCKTYPE must be defined at compile time using the -DLOCKTYPE=<lockname>"
#else
typedef ppp::LOCKTYPE Lock;
#endif

// Data arrays
static Lock* lock_array;
static int* counter_array;
static int* counter_check_array;


void driver()  // Called for each thread
{

  if (ppp::get_thread_id() == 0) {

    // Allocation 
    lock_array = new Lock[num_locks];
    counter_array = new int[num_counters];
    counter_check_array = new int[num_counters];
    
    for (int i = 0; i<num_counters; i++) {
      counter_array[i] = 0;
      counter_check_array[i] = 0;
    }
  }

  int my_iterations = num_iterations / ppp::get_thread_count();

  double start_time = get_seconds();
  int counters_per_lock = num_counters / num_locks;
    
  ppp::barrier();

  // do the work
  for (int iteration = 0; iteration < my_iterations; iteration++) {

    // Think time
    for (int i = 0; i<think_time; i++) {
      ppp::compiler_fence();
    }

    int lock_id = ppp::threadsafe_random() % num_locks;
    int counter_id = ppp::threadsafe_random() % counters_per_lock;
    int counter_index = ((lock_id * counters_per_lock) + counter_id);

    lock_array[lock_id].acquire();
    counter_array[counter_index]++;

    // Hold time
    for (int i = 0; i<hold_time; i++) {
      ppp::compiler_fence();
    }

    lock_array[lock_id].release();
  }
  
  ppp::barrier();
    
  double end_time = get_seconds();
  double seconds = end_time - start_time;

  // Check the result
  ppp::reseed_random_per_thread();  // reset the random number generator

  check_lock.acquire();
  for (int iteration = 0; iteration < my_iterations; iteration++) {
    int lock_id = ppp::threadsafe_random() % num_locks;
    int counter_id = ppp::threadsafe_random() % counters_per_lock;
    int counter_index = ((lock_id * counters_per_lock) + counter_id);
    counter_check_array[counter_index]++;
  }
  check_lock.release();

  ppp::barrier();

  // Check the output
  if (ppp::get_thread_id() == 0) {
    for (int i=0; i<num_counters; i++) {
      if (counter_check_array[i] != counter_array[i]) {
        fprintf(stderr, "ERROR: At element %d, counter_check_array (%d) and counter_array (%d) don't match\n", i, counter_check_array[i], counter_array[i]);
      }
    }
  }
    
  // Display timing
  if (ppp::get_thread_id() == 0) {
    std::cout << "user: " << getuser() << ", ";
    std::cout << "host: " << gethost() << ", ";
    std::cout << "threads: " << num_threads << ", ";
    std::cout << "locks: " << num_locks << ", ";
    std::cout << "counters: " << num_counters << ", ";
    std::cout << "iterations: " << num_iterations << ", ";
    std::cout << "seconds: " << seconds << "\n";
  }
}

static struct option long_options[] =
  {
    {"counters", required_argument, 0, 'c'},
    {"locks", required_argument, 0, 'l'},
    {"iterations", required_argument, 0, 'i'},
    {"threads", required_argument, 0, 'n'},
    {"holdtime", required_argument, 0, 'h'},
    {"thinktime", required_argument, 0, 't'},
    {0, 0, 0, 0}
  };

int main(int argc, char** argv)
{
  
  // Code based on the example code in the getopt documentation
  while (true) {

    int option_index = 0;    
    int c = getopt_long_only(argc, argv, "c:l:i:n:h:t:",
                             long_options, &option_index);
    
    /* Detect the end of the options. */
    if (c == -1)
      break;
    
    switch (c) {
    case 0:
      /* If this option set a flag, do nothing else now. */
      break;
      
    case 'c':
      num_counters = atoi(optarg);
      break;
      
    case 'l':
      num_locks = atoi(optarg);
      break;

    case 'i':
      num_iterations = atoi(optarg);
      break;
      
    case 'n':
      num_threads = atoi(optarg);
      break;

    case 'h':
      hold_time = atoi(optarg);
      break;
      
    case 't':
      think_time = atoi(optarg);
      break;

    case '?':
      /* getopt_long already printed an error message. */
      exit(1);
      break;
      
    default:
      exit(1);
    }
  }

  if (num_locks <= 0) {
    fprintf(stderr, "Error: --locks must be set\n");
    exit(1);
  }

  if (num_iterations <= 0) {
    fprintf(stderr, "Error: --iterations must be set\n");
    exit(1);
  }

  if (num_locks > num_counters) {
    fprintf(stderr, "Number of locks must be less than or equal the number of counters\n");
    exit(1);
  }
  
  ppp::init(driver, num_threads);
}

